from decouple import config

MANIFEST_PATH = config("MANIFEST_PATH", default="inputs/manifest.json")
CATALOG_PATH = config("CATALOG_PATH", default="inputs/catalog.json")
SCHEMA_MAPPING_PATH = config("SCHEMA_MAPPING_PATH", default="outputs/schema_mapping.json")
MANIFEST_CL_PATH = config("MANIFEST_CL_PATH", default="outputs/manifest_column_lineage.json")

DIALECT = config("DIALECT", default="snowflake")

UNKNOWN = "UNKNOWN"



from typing import Callable, Iterable

import networkx as nx
from rapidfuzz.fuzz import partial_ratio
from rapidfuzz.process import extract

SCORER = partial_ratio  # no preprocessing is applied e.g. whitespace trimming or case normalization
SCORE_CUTOFF = 70

NODE_RELEVANT_ATTRIBUTES = [
    "description",
    "database",
    "schema",
    "alias",
    "relation_name",
    "resource_type",
    "path",
    "unique_id",
    "tags",
    "metrics",
    "depends_on",
    "columns",
    "compiled_code",
]
RELEVENT_ATTRIBUTES = {
    "model": NODE_RELEVANT_ATTRIBUTES,
    "source": [
        "description",
        "database",
        "schema",
        "identifier",
        "relation_name",
        "resource_type",
        "path",
        "unique_id",
        "tags",
        "columns",
    ],
    "exposure": ["description", "resource_type", "type", "owner", "path", "unique_id", "tags", "metrics", "depends_on"],
    "test": NODE_RELEVANT_ATTRIBUTES,
    "seed": NODE_RELEVANT_ATTRIBUTES,
    "operation": NODE_RELEVANT_ATTRIBUTES,
    "snapshot": NODE_RELEVANT_ATTRIBUTES,
}


def get_useful_attributes(dbt_unique_id: str, G: nx.DiGraph):
    attributes = G.nodes[dbt_unique_id]
    # Return everything if resource_type not one of the main ones
    if attributes["resource_type"] not in RELEVENT_ATTRIBUTES:
        return attributes
    return {key: attributes[key] for key in RELEVENT_ATTRIBUTES[attributes["resource_type"]]}


def get_useful_attributes_for_models(dbt_unique_ids: Iterable[str], G: nx.DiGraph):
    return {dbt_unique_id: get_useful_attributes(dbt_unique_id, G) for dbt_unique_id in dbt_unique_ids}


def get_dbt_node_attributes_tool(G: nx.DiGraph):
    def get_dbt_node_attributes(dbt_unique_id: str):
        """Get attributes of a given node.

        Args:
            dbt_unique_id (str): The dbt unique id of the node to get attributes for

        Returns:
            dict: Dictionary containing relevant attributes of the node
        """
        return get_useful_attributes(dbt_unique_id, G)

    return get_dbt_node_attributes


def get_dbt_predecessor_tool(G: nx.DiGraph):
    def get_dbt_predecessors(dbt_unique_id: str):
        """Get the predecessor nodes of a given node. Returns a list of dbt unique ids.

        To get a node's attributes, use the get_dbt_node_attributes tool. Predecessors are nodes that are directly
        upstream of the given node (i.e. only one edge away).

        Args:
            dbt_unique_id (str): The dbt unique id of the node to get predecessors for

        Returns:
            list: List of dbt unique ids. To get their attributes, use the get_dbt_node_attributes tool.
        """
        predecessor_nodes = list(G.predecessors(dbt_unique_id))
        return predecessor_nodes

    return get_dbt_predecessors


def get_dbt_successor_tool(G: nx.DiGraph):
    def get_dbt_successors(dbt_unique_id: str):
        """Get the successor nodes of a given node. Returns a list of dbt unique ids.

        To get a node's attributes, use the get_dbt_node_attributes tool. Successors are nodes that are directly
        downstream of the given node (i.e. only one edge away).

        Args:
            dbt_unique_id (str): The dbt unique id of the node to get successors for

        Returns:
            list: List of dbt unique ids. To get their attributes, use the get_dbt_node_attributes tool.
        """
        successor_nodes = list(G.successors(dbt_unique_id))
        return successor_nodes

    return get_dbt_successors


def get_dbt_node_attributes_from_columns(columns: list[str], G: nx.DiGraph):
    model_attributes_for_columns = {}
    for column in columns:
        dbt_unique_id, _ = column.rsplit(".", 1)
        model_attributes_for_columns[column] = get_useful_attributes(dbt_unique_id, G)
    return model_attributes_for_columns


def get_column_ancestors_tool(G: nx.DiGraph, G_col: nx.DiGraph):
    def get_column_ancestors(dbt_unique_id: str, column_name: str) -> list[dict]:
        """Get the ancestor models for a specific column.

        Ancestors are all models that the specified column depends on, even ones that are more than one edge away.
        As such, it is a much more targeted tool and should be used over get_model_predecessors if you are interested
        in a specific column/field.

        Args:
            model_name (str): The name of the model containing the column
            column_name (str): The name of the column to get ancestors for

        Returns:
            dict: mapping of ancestor column keys in format {dbt_unique_id.column_name} to that model's attributes
        """
        ancestor_columns = nx.ancestors(G_col, f"{dbt_unique_id}.{column_name}")
        return get_dbt_node_attributes_from_columns(ancestor_columns, G)

    return get_column_ancestors


def get_column_descendants_tool(G: nx.DiGraph, G_col: nx.DiGraph):
    def get_column_descendants(dbt_unique_id: str, column_name: str):
        """Get the descendant models for a specific column.

        Descendants are all models that depend on the specified column, even ones that are more than one edge away.
        As such, it is a much more targeted tool and should be used over get_model_successors if you are interested
        in a specific column/field.

        Args:
            model_name (str): The name of the model containing the column
            column_name (str): The name of the column to get descendants for

        Returns:
            dict: mapping of ancestor column keys in format {dbt_unique_id.column_name} to that model's attributes
        """
        descendant_columns = nx.descendants(G_col, f"{dbt_unique_id}.{column_name}")
        return get_dbt_node_attributes_from_columns(descendant_columns, G)

    return get_column_descendants


def model_names_from_matches(matches: list[tuple[str, float, str]]):
    return [match[2] for match in matches]


def get_dbt_node_search_tool(G: nx.DiGraph):
    dbt_unique_ids = {node: node for node in list(G.nodes.keys())}

    def search_dbt_node_names(search_string: str, first_n: int = 10):
        """Search for dbt nodes by name. Returns a list of dbt unique ids.

        To get their attributes, use the get_dbt_node_attributes tool.
        Args:
            search_string (str): The string to search for in model names
            first_n (int, optional): Maximum number of results to return. Defaults to 10.

        Returns:
            list: List of dbt unique ids. To get their attributes, use the get_dbt_node_attributes tool.
        """
        matches = extract(
            search_string,
            dbt_unique_ids,
            limit=first_n,
            scorer=SCORER,
            score_cutoff=SCORE_CUTOFF,
        )
        matching_models = model_names_from_matches(matches)
        return matching_models

    return search_dbt_node_names


def get_dbt_column_search_tool(G: nx.DiGraph):
    column_names = {node_name: column for node_name, node in G.nodes.items() for column in node.get("columns", [])}

    def search_dbt_column_names(search_string: str, first_n: int = 10):
        """Search for nodes containing columns matching a name pattern. Returns a list of dbt unique ids.

        To get their attributes, use the get_dbt_node_attributes tool.
        Args:
            search_string (str): The string to search for in column names
            first_n (int, optional): Maximum number of models to return. Defaults to 10.

        Returns:
            list: List of dbt unique ids. To get their attributes, use the get_dbt_node_attributes tool.

        """
        matches = extract(
            search_string,
            column_names,
            limit=first_n,
            scorer=SCORER,
            score_cutoff=SCORE_CUTOFF,
        )
        matching_models = model_names_from_matches(matches)
        return matching_models

    return search_dbt_column_names


def get_dbt_sql_search_tool(G: nx.DiGraph):
    compiled_code = {node_name: node.get("compiled_code", "") for node_name, node in list(G.nodes.items())}

    def search_dbt_sql_code(search_string: str, first_n: int = 10):
        """Search for nodes containing a string in their SQL code. Returns a list of dbt unique ids.

        To get their attributes, use the get_dbt_node_attributes tool.
        Args:
            search_string (str): The string to search for in SQL code
            first_n (int, optional): Maximum number of results to return. Defaults to 10.

        Returns:
            list: List of dbt unique ids. To get their attributes, use the get_dbt_node_attributes tool.

        """
        matches = extract(search_string, compiled_code, limit=first_n, scorer=SCORER)
        matching_models = model_names_from_matches(matches)
        return matching_models

    return search_dbt_sql_code


def get_dbt_tools(G: nx.DiGraph, G_col: nx.DiGraph = None) -> list[Callable]:
    """Get all tools for working with the DBT graphs.

    Args:
        manifest_path (str): The path to the manifest file
        schema_mapping_path (str): The path to the schema mapping file
        manifest_cl_path (str): The path to the manifest column lineage file

    Returns:
        list: List of tool functions for working with DBT data
    """
    tools = [
        get_dbt_predecessor_tool(G=G),
        get_dbt_successor_tool(G=G),
        get_dbt_node_attributes_tool(G=G),
        get_dbt_node_search_tool(G=G),
        get_dbt_column_search_tool(G=G),
        get_dbt_sql_search_tool(G=G),
    ]
    if G_col is not None:
        tools.extend(
            [
                get_column_ancestors_tool(G=G, G_col=G_col),
                get_column_descendants_tool(G=G, G_col=G_col),
            ]
        )
    return tools



import json

from dbt.artifacts.schemas.catalog import CatalogArtifact
from dbt.artifacts.schemas.manifest import WritableManifest

from dbt_docs_mcp.constants import CATALOG_PATH, MANIFEST_PATH


def write_json(json_data, file_path):
    with open(file_path, "w") as f:
        json.dump(json_data, f, indent=4)


def read_json(file_path):
    with open(file_path, "r") as f:
        return json.load(f)


def load_manifest(manifest_path: str = MANIFEST_PATH):
    # Read the manifest JSON first
    with open(manifest_path, "r") as f:
        manifest_json = json.load(f)
    # Patch missing 'unit_tests' field if necessary
    if "unit_tests" not in manifest_json:
        manifest_json["unit_tests"] = {}
    # Use WritableManifest.from_dict instead of .read
    manifest = WritableManifest.from_dict(manifest_json)
    return manifest

def load_catalog(catalog_path: str = CATALOG_PATH):
    catalog = CatalogArtifact.read(catalog_path)
    return catalog



import logging
from pathlib import Path

from mcp.server.fastmcp import FastMCP

from dbt_docs_mcp.constants import (
    CATALOG_PATH,  # Assuming CATALOG_PATH is defined here
    MANIFEST_CL_PATH,
    MANIFEST_PATH,
    SCHEMA_MAPPING_PATH,
)
from dbt_docs_mcp.dbt_graphs import get_G_and_G_col
from dbt_docs_mcp.dbt_processing import read_or_write_schema_mapping
from dbt_docs_mcp.tools import get_dbt_tools
from dbt_docs_mcp.utils import load_catalog, load_manifest

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)

"""
DBT Docs MCP Server: A tool for exploring DBT documentation and lineage.

If you have not yet created a manifest_column_lineage.json file or a schema_mapping.json file,
you can use the following command: python scripts/create_manifest_cl.py
If you have (with standard defaults) then you can simply run:
mcp run mcp_server.py

Handles different scenarios for finding input files:
1. manifest.json, schema_mapping.json, manifest_column_lineage.json exist -> Load full graphs.
2. manifest.json, schema_mapping.json exist (no col lineage) -> Load graph G, G_col=None.
3. manifest.json, catalog.json exist (no schema mapping/col lineage) -> Create schema_mapping.json, then load G, G_col=None.

Run command: mcp run mcp_server.py
"""  # noqa: E501

mcp = FastMCP("DBT Docs")

# --- File Loading Logic ---

# 1. Check essential MANIFEST_PATH first
logger.info(f"Looking for manifest file at: {MANIFEST_PATH}")
if not Path(MANIFEST_PATH).exists():
    raise FileNotFoundError(f"Required file not found: {MANIFEST_PATH}. Cannot proceed.")
logger.info("Manifest file found.")

schema_mapping_exists = Path(SCHEMA_MAPPING_PATH).exists()
manifest_cl_path_to_use = None  # Default to no column lineage

if schema_mapping_exists:
    logger.info(f"Found schema mapping file: {SCHEMA_MAPPING_PATH}")
    # Scenario 1 or 2 depends on manifest_cl_path
    if Path(MANIFEST_CL_PATH).exists():
        logger.info(f"Found column lineage file: {MANIFEST_CL_PATH}. Will load column lineage graph.")
        manifest_cl_path_to_use = MANIFEST_CL_PATH
    else:
        logger.warning(f"Column lineage file not found at: {MANIFEST_CL_PATH}. Proceeding without column lineage.")
else:
    # Schema mapping not found, check if we can create it (Scenario 3)
    logger.warning(f"Schema mapping file not found: {SCHEMA_MAPPING_PATH}")
    logger.info(f"Looking for catalog file at: {CATALOG_PATH}")
    if Path(CATALOG_PATH).exists():
        logger.info(f"Found catalog file: {CATALOG_PATH}. Attempting to create schema mapping...")
        try:
            manifest = load_manifest(MANIFEST_PATH)
            catalog = load_catalog(CATALOG_PATH)

            schema_mapping = read_or_write_schema_mapping(manifest, catalog, SCHEMA_MAPPING_PATH)

            if not Path(SCHEMA_MAPPING_PATH).exists():
                raise RuntimeError(f"Utility function failed to create schema mapping file at {SCHEMA_MAPPING_PATH}.")
            logger.info(f"Successfully created schema mapping file: {SCHEMA_MAPPING_PATH}")
            # Proceed without column lineage as it wasn't present initially
            logger.warning(
                f"Column lineage file ({MANIFEST_CL_PATH}) was not found. Proceeding without column lineage."
            )
        except Exception as e:
            logger.error(f"Error creating schema mapping file from catalog: {e}", exc_info=True)
            raise RuntimeError(f"Failed to create schema mapping file from catalog: {e}")
    else:
        # Error: Cannot proceed without schema mapping or catalog
        raise FileNotFoundError(
            f"Required file not found: {SCHEMA_MAPPING_PATH}."
            f"Also could not find catalog file to generate it: {CATALOG_PATH}."
            "Please provide a valid schema_mapping.json or ensure manifest.json and catalog.json are present."
        )

# --- Load Graphs ---
logger.info("Loading dbt graph data...")
G, G_col = get_G_and_G_col(
    manifest_path=MANIFEST_PATH,
    schema_mapping_path=SCHEMA_MAPPING_PATH,  # Should exist now or error was raised
    manifest_cl_path=manifest_cl_path_to_use,
)
logger.info("Graph loading complete.")
if G_col is None:
    logger.info("Note: Column lineage graph (G_col) was not loaded.")
else:
    logger.info("Column lineage graph (G_col) loaded successfully.")


# --- Setup MCP Server ---
mcp = FastMCP("DBT Docs")

tools = get_dbt_tools(G, G_col)  # Pass potentially None G_col

logger.info(f"Adding {len(tools)} tools to MCP server.")
for tool in tools:
    mcp.add_tool(tool)

if __name__ == "__main__":
    logger.info("Starting MCP server...")
    mcp.run()



import warnings

import networkx as nx
from dbt.artifacts.resources import ColumnInfo
from dbt.artifacts.schemas.catalog import CatalogKey
from dbt.artifacts.schemas.manifest import WritableManifest
from dbt.task.docs.generate import get_unique_id_mapping

from dbt_docs_mcp.constants import UNKNOWN
from dbt_docs_mcp.utils import load_manifest, read_json


def get_dbt_graph(manifest: WritableManifest, schema: dict) -> nx.DiGraph:
    """Get the dbt graph from a manifest.

    The graph includes nodes, sources and exposures.
    """
    G = nx.DiGraph()
    nodes, edges = [], []
    for k, v in {**manifest.nodes, **manifest.sources, **manifest.exposures}.items():
        if v.resource_type != "exposure":
            alias = getattr(v, "alias", None)
            identifier = getattr(v, "identifier", None)
            table = alias or identifier
            try:
                v.columns = {
                    column_name: ColumnInfo(name=column_name, data_type=data_type)
                    for column_name, data_type in schema[v.database.lower()][v.schema.lower()][table].items()
                }
            except KeyError:
                warnings.warn(
                    f"Node {k} cannot be found in the schema with path: {v.database.lower()}.{v.schema.lower()}.{table}"
                )

        nodes.append((k, vars(v)))
        depends_on_nodes = getattr(getattr(v, "depends_on", {}), "nodes", [])
        edges.extend((d, k) for d in depends_on_nodes)
    G.add_nodes_from(nodes)
    G.add_edges_from(edges)
    return G


def get_column_lineage_graph(manifest_column_lineage, node_map, source_map) -> nx.DiGraph:
    """Get the column lineage graph from a manifest and a node map and source map.

    Edges represent whether a column depends on another column.
    """
    G_col = nx.DiGraph()
    nodes, edges = [], []
    for model_unique_id, table_column_lineage in manifest_column_lineage.items():
        for column_name, column_lineage in table_column_lineage.items():
            column_unique_id = f"{model_unique_id}.{column_name}"
            nodes.append((column_unique_id, {}))
            for parent_column in column_lineage:
                if parent_column["database_fqn"] == UNKNOWN:
                    continue
                parent_catalog_key = CatalogKey(*parent_column["database_fqn"].split("."))
                if parent_catalog_key not in node_map and parent_catalog_key not in source_map:
                    warnings.warn(
                        f"Model {model_unique_id} has a column {column_name} with a parent "
                        f"{parent_column['database_fqn']} that is not in the node_map or source_map"
                    )
                    continue
                parent_dbt_unique_id = node_map.get(parent_catalog_key) or list(source_map.get(parent_catalog_key))[0]
                parent_unique_id = f"{parent_dbt_unique_id}.{parent_column['column_name']}"
                edges += [(parent_unique_id, column_unique_id)]
    G_col.add_nodes_from(nodes)
    G_col.add_edges_from(edges)
    return G_col


def get_G_and_G_col(
    manifest_path: str,
    schema_mapping_path: str,
    manifest_cl_path: str,
) -> tuple[nx.DiGraph, nx.DiGraph]:
    manifest = load_manifest(manifest_path)
    schema_mapping = read_json(schema_mapping_path)
    G = get_dbt_graph(manifest=manifest, schema=schema_mapping)

    if manifest_cl_path:
        manifest_column_lineage = read_json(manifest_cl_path)
        node_map, source_map = get_unique_id_mapping(manifest)
        G_col = get_column_lineage_graph(
            manifest_column_lineage=manifest_column_lineage, node_map=node_map, source_map=source_map
        )
    else:
        G_col = None

    return G, G_col



import warnings
from collections import defaultdict
from pathlib import Path

from dbt.artifacts.schemas.catalog import CatalogArtifact
from dbt.artifacts.schemas.manifest import WritableManifest
from sqlglot import expressions as exp
from sqlglot import parse_one
from sqlglot.errors import OptimizeError, SqlglotError
from sqlglot.lineage import Node, lineage
from sqlglot.optimizer.qualify import qualify
from tqdm import tqdm

from dbt_docs_mcp.constants import DIALECT, UNKNOWN
from dbt_docs_mcp.utils import read_json, write_json


def create_database_schema_table_mapping_from_sql(manifest: WritableManifest, schema: dict = {}) -> dict:
    """Supplement an existing schema with columns from models that don't have a catalog entry.

    If the schema already contains columns for a model, it will not change it.
    Args:
        manifest: A WritableManifest object containing the manifest data.
        schema: A dictionary mapping from databases to schemas to tables to columns.

    Returns:
        A dictionary mapping from databases to schemas to tables to columns.
    """
    for model in tqdm(manifest.nodes.values()):
        if (
            schema.get(model.database.lower(), {}).get(model.schema.lower(), {}).get(model.name.lower(), None)
            is not None
        ):
            continue
        expression = parse_one(model.compiled_code, dialect=DIALECT)
        try:
            qualified_expression = qualify(
                expression,
                schema=schema,
                validate_qualify_columns=False,
                allow_partial_qualification=False,
                dialect=DIALECT,
            )
        except OptimizeError as e:
            warnings.warn(
                f"Error qualifying columns for {model.unique_id}: {e}. Performing partial",
                "qualification without validation.",
                UserWarning,
            )
            qualified_expression = qualify(
                expression,
                schema=schema,
                validate_qualify_columns=True,
                allow_partial_qualification=True,
                dialect=DIALECT,
            )
        schema[model.database.lower()][model.schema.lower()][model.name.lower()] = {
            name.lower(): None for name in qualified_expression.named_selects
        }

    return schema


def create_database_schema_table_column_mapping(manifest: WritableManifest, catalog: CatalogArtifact) -> dict:
    """Returns a nested dictionary mapping from databases to schemas to tables to columns, for both nodes and sources.

    Args:
        manifest: A WritableManifest object containing the manifest data.
        catalog: A CatalogArtifact object containing the catalog data.

    Returns:
        A nested dictionary mapping from databases to schemas to tables to columns.
    """
    result = defaultdict(lambda: defaultdict(dict))

    # Process both nodes and sources which have the same structure
    for collection in [catalog.nodes, catalog.sources]:
        for _, table in collection.items():
            database = table.metadata.database.lower() or ""
            schema = table.metadata.schema.lower()
            table_name = table.metadata.name.lower()

            if database not in result:
                result[database] = defaultdict(dict)
            if schema not in result[database]:
                result[database][schema] = {}

            # Add columns to the mapping
            result[database][schema][table_name] = {k.lower(): v.type.lower() for k, v in table.columns.items()}

    # Some table aren't in the catalog e.g. ephemeral tables (and others not sure why?)
    result = create_database_schema_table_mapping_from_sql(manifest, result)

    return result


def get_parent_nodes_from_lineage_node(node: Node) -> list[Node]:
    """Get the parent nodes from a lineage node.

    Args:
        node: A lineage node.

    Returns:
        A list of parent nodes.
    """
    parent_nodes = []
    for int_node in node.walk():
        if isinstance(int_node.source, exp.Table):
            parent_nodes.append(int_node)
    return parent_nodes


def get_column_lineage(column_name: str, sql: str, schema: dict, dialect: str = DIALECT) -> list[dict]:
    """Get the lineage of a column.

    Args:
        column_name: The name of the column.
        sql: The SQL code of the model.
        schema: The schema of the model.
        dialect: The dialect of the SQL code.

    Returns:
        A list of parent nodes with their column_name and database_fqn.
    """
    lineage_node = lineage(column_name, sql=sql, schema=schema, dialect=dialect)
    parent_nodes = get_parent_nodes_from_lineage_node(lineage_node)
    column_lineage = []
    for parent_node in parent_nodes:
        column_lineage.append(
            {
                "column_name": parent_node.name.split(".")[-1].lower(),
                "database_fqn": (
                    f"{parent_node.source.catalog}.{parent_node.source.db}.{parent_node.source.this}".lower()
                ),
            }
        )
    return column_lineage


def get_column_lineage_for_model(model, schema: dict, dialect: str = DIALECT) -> dict[str, list[dict]]:
    sql = model.compiled_code
    table_column_lineage = {}
    for column_name in schema[model.database.lower()][model.schema.lower()][model.name.lower()]:
        try:
            table_column_lineage[column_name] = get_column_lineage(column_name, sql, schema, dialect)
        except SqlglotError as e:
            warnings.warn(f"Error getting column lineage for model {model.unique_id}, column {column_name}: {e}")
            table_column_lineage[column_name] = [{"column_name": UNKNOWN, "database_fqn": UNKNOWN}]
    return table_column_lineage


def get_column_lineage_for_manifest(
    manifest: WritableManifest, schema: dict, dialect: str = DIALECT
) -> dict[str, dict[str, list[dict]]]:
    manifest_column_lineage = {}
    for model_unique_id, model in tqdm(manifest.nodes.items()):
        if model.resource_type not in ["model", "test"]:
            continue
        table_column_lineage = get_column_lineage_for_model(model, schema, dialect)
        manifest_column_lineage[model_unique_id] = table_column_lineage
    return manifest_column_lineage


def read_or_write_schema_mapping(
    manifest: WritableManifest, catalog: CatalogArtifact, schema_mapping_path: str, overwrite: bool = False
) -> dict:
    # Ensure output directories exist
    Path(schema_mapping_path).parent.mkdir(parents=True, exist_ok=True)

    if Path(schema_mapping_path).exists() and not overwrite:
        schema_mapping = read_json(schema_mapping_path)
    else:
        # Create database schema table column mapping
        print("Creating database schema table column mapping...")
        schema_mapping = create_database_schema_table_column_mapping(manifest, catalog)

        print(f"Saving results to {schema_mapping_path}...")
        write_json(schema_mapping, schema_mapping_path)

    return schema_mapping


def read_or_write_manifest_column_lineage(
    manifest: WritableManifest, schema_mapping: dict, manifest_cl_path: str, overwrite: bool = False
) -> dict:
    # Ensure output directories exist
    Path(manifest_cl_path).parent.mkdir(parents=True, exist_ok=True)

    if Path(manifest_cl_path).exists() and not overwrite:
        manifest_cl = read_json(manifest_cl_path)
    else:
        # Generate column-level lineage for the entire manifest
        print("Generating column-level lineage...")
        manifest_cl = get_column_lineage_for_manifest(manifest=manifest, schema=schema_mapping, dialect=DIALECT)

        print(f"Saving results to {manifest_cl_path}...")
        write_json(manifest_cl, manifest_cl_path)

    return manifest_cl
